"""All delatore outputs"""
from threading import Event

from apubsub import Service

from .telegram import BotRunner


async def start_outputs(service: Service, stop_event: Event):
    await BotRunner(service, stop_event).start()



"""Base Source implementation"""

import asyncio
import logging
from abc import ABC, ABCMeta, abstractmethod
from inspect import isabstract
from typing import Optional, Union

from apubsub.client import Client

from delatore.configuration import SOURCES_CFG
from delatore.configuration.static import SourceConfiguration

Json = Union[dict, list]

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


class NoUpdates(Exception):
    """Raised when source has no updates"""


class SourceMeta(ABCMeta):
    """Source class meta setting up data from configuration"""

    CONFIG_ID: str
    TOPIC: str
    config: SourceConfiguration

    def __init__(cls, name, bases=(), dct=None):  # pylint:disable=redefined-builtin
        super().__init__(name, bases, dct)
        if not isabstract(cls):  # skip configuration for abstracts
            assert hasattr(cls, 'CONFIG_ID')
            cls.config = SOURCES_CFG[cls.CONFIG_ID]
            cls.TOPIC = cls.config.publishes


class Source(ABC, metaclass=SourceMeta):
    """Source API posting updates to """

    CONFIG_ID: str = NotImplemented

    client: Client
    config: SourceConfiguration

    def __init__(self, client: Client,
                 polling_interval=10.0,
                 request_timeout=10.0,
                 ignore_duplicates=True):
        self.client = client
        self.polling_interval = polling_interval
        self.request_timeout = request_timeout
        self.ignore_duplicates = ignore_duplicates

    @abstractmethod
    async def get_update(self) -> Optional[Json]:
        """Get source update"""

    @classmethod
    @abstractmethod
    def convert(cls, data: Json) -> str:
        """Converts json-like object to Telegram supported Markdown

        :param data: JSON dictionary or list
        :return: Markdown string
        """

    async def start(self, stop_event: asyncio.Event):
        """Start processing updates"""
        last = None
        name = type(self).__name__
        LOGGER.info('Source %s started', name)
        while not stop_event.is_set():
            try:
                new = await asyncio.wait_for(self.get_update(), self.request_timeout)
            except (asyncio.TimeoutError, NoUpdates):
                continue
            if new in [last, None] and self.ignore_duplicates:
                continue
            LOGGER.debug('New data received from source: %s\ndata:\n%s', name, new)
            md_message = self.convert(new)
            await self.client.publish(self.TOPIC, md_message)
            last = new
            LOGGER.debug('Wait for new data for %s', self.polling_interval)
            await asyncio.sleep(self.polling_interval)



"""Configurations fixed for library"""
from dataclasses import dataclass
from typing import Dict, List, Type, TypeVar

import yaml
from ocomone import Resources


@dataclass(frozen=True)
class Configuration:
    """Base configuration class"""
    name: str
    params: dict


@dataclass(frozen=True)
class OutputConfiguration(Configuration):
    """Single output service configuration"""
    subscriptions: List[str]


@dataclass(frozen=True)
class SourceConfiguration(Configuration):
    """Single source service configuration"""
    publishes: str


_CONFIGS = Resources(__file__)
_SOURCES_CFG_FILE = _CONFIGS['sources.yaml']
_OUTPUTS_CFG_FILE = _CONFIGS['outputs.yaml']

T = TypeVar('T', SourceConfiguration, OutputConfiguration)


def _load_configuration(cfg_file: str, cfg_class: Type[T]) -> Dict[str, T]:
    with open(cfg_file, 'r') as src_cfg:
        configs = yaml.safe_load(src_cfg)
    result = {cfg['name']: cfg_class(**cfg) for cfg in configs}
    return result


SOURCES_CFG = _load_configuration(_SOURCES_CFG_FILE, SourceConfiguration)
OUTPUTS_CFG = _load_configuration(_OUTPUTS_CFG_FILE, OutputConfiguration)



import asyncio
import logging

from apubsub import Service

from .awx_api import AWXApiSource
from .http import AWXWebHookSource
from .influx import InfluxSource

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


async def start_sources(msg_service: Service, stop_event: asyncio.Event):
    """Start all sources and make them report to message queue"""
    await asyncio.wait([
        AWXApiSource(msg_service.get_client()).start(stop_event),
        AWXWebHookSource(msg_service.get_client()).start(stop_event),
        InfluxSource(msg_service.get_client()).start(stop_event),
    ])



import logging
from abc import ABC
from asyncio import Queue

from aiohttp import web
from aiohttp.web_request import Request
from apubsub.client import Client

from .base import Source
from ..emoji import Emoji, replace_emoji
from ..json2mdwn import convert

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


async def ok(_: Request):
    """Send `OK` response"""
    return web.Response(text='OK')


class HttpListenerSource(Source, ABC):
    """HTTP listener"""

    # pylint: disable=abstract-method

    def __init__(self, client: Client, polling_interval=10, request_timeout=10.0):
        super().__init__(client,
                         ignore_duplicates=True,
                         polling_interval=polling_interval,
                         request_timeout=request_timeout)
        app = web.Application()
        app.add_routes([web.get('/', ok)])
        self.app = app
        self.port = self.config.params['port']
        self.runner = web.AppRunner(self.app)
        self.site = None

    async def start(self, stop_event):
        """Start server coroutine"""
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, port=self.port)
        await self.site.start()
        LOGGER.debug('Web server started on port %s', self.port)
        await super().start(stop_event)
        await self.runner.cleanup()

    async def stop(self):
        await self.site.stop()


AWX_LISTENER_EMOJI_MAP = {
    '`failed`': Emoji.FAILED,
    '`running`': Emoji.RUNNING,
    '`success`': Emoji.SUCCESS,
    '`cancelled`': Emoji.CANCELED
}


class AWXWebHookSource(HttpListenerSource):
    """HTTP listener for AWX web hooks"""

    CONFIG_ID = 'awx_web_hook'

    @classmethod
    def convert(cls, data: dict) -> str:
        data.pop('From', None)
        text = convert(data).strip()
        text = '* From Ansible Tower *\n' + replace_emoji(text, AWX_LISTENER_EMOJI_MAP)
        return text

    def __init__(self, client):
        super().__init__(client,
                         polling_interval=10,  # checking input queue interval
                         request_timeout=.1)  # checking input queue timeout
        self.app.add_routes([
            web.post('/notifications', self.notifications)
        ])
        self.updates = Queue()

    async def notifications(self, request: web.Request):
        await self.updates.put(await request.json())
        return web.Response(text='OK')

    async def get_update(self):
        return await self.updates.get()



"""Conversion of source data to markdown v2"""

import json
import re


def normalize_key(src: str):
    """Normalize dictionary key format"""
    if src.isupper():
        return f'`{src}`'
    return src.replace('_', ' ').title()


def backtick_values(data):
    """Escape json values with backticks

    To define your behaviour, implement `__md__` method in your class
    """
    if hasattr(data, '__md__'):
        return data.__md__()
    if isinstance(data, dict):
        return {f'{normalize_key(key)} ': backtick_values(data[key]) for key in data}
    if isinstance(data, list):
        return [backtick_values(item) for item in data]
    return f'`{data}`'


__MARKS = re.compile(r'[{}",\[\]]')
__MD2_SPECIAL = re.compile(r'([()])')


def telegram_escape(text):
    """Escape telegram message text"""
    return __MD2_SPECIAL.sub(r'\\\1', text)  # MarkdownV2 chars are escaped


def convert(data):
    """Convert dict or list to nice-looking telegram markdown"""
    formatted = __MARKS.sub('', json.dumps(backtick_values(data),
                                           separators=('', ':    '),
                                           indent=4,
                                           ensure_ascii=False))
    formatted = telegram_escape(formatted)
    formatted = '\n'.join(line[4:].rstrip() for line in formatted.split('\n') if line.rstrip() != '')
    return formatted



import asyncio
import logging
import time
from typing import List, NamedTuple, Optional

from aiohttp import ClientSession
from apubsub.client import Client

from .base import NoUpdates, Source
from ..configuration import BOT_CONFIG
from ..emoji import Emoji, replace_emoji
from ..json2mdwn import convert

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

AWX_EMOJI_MAP = {
    'successful': Emoji.SUCCESS,
    'failed': Emoji.FAILED,
    'never updated': Emoji.NO_DATA,
}

TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'

NO_TEMPLATE_PATTERN = 'No template with name \'{}\' found'  # pylint:disable=invalid-string-quote


class TemplateStatus(NamedTuple):
    name: str
    last_run_timestamp: str
    last_status: str
    playbook: str

    def __md__(self):
        status = replace_emoji(self.last_status, AWX_EMOJI_MAP, '%e')
        timestamp = self.last_run_timestamp
        if timestamp is not None:
            timestamp = time.strftime('%d.%m.%y %H:%M', time.strptime(timestamp, TIMESTAMP_FORMAT))
        return rf'{status}   —   `{self.name}`  (`{timestamp}`)'


def get_session():
    """Get authorized session instance"""
    return ClientSession(headers={'Authorization': f'Bearer {BOT_CONFIG.awx_auth_token}'})


class AWXParams(NamedTuple):
    """Influx params storage"""
    host: str
    topic_in: str


class AWXApiSource(Source):
    """AWX API Client"""

    CONFIG_ID = 'awx_api'
    _params: Optional[AWXParams] = None

    async def get_update(self):
        template_name = await self.client.get(self.polling_interval)
        if template_name is None:
            raise NoUpdates

        _filter = single_template_filter(template_name)
        data = await self.get_templates(_filter) or [NO_TEMPLATE_PATTERN.format(template_name)]
        return data

    @classmethod
    def convert(cls, data: list) -> str:
        return f'* AWX scenarios status: *\n{convert(data)}'

    @classmethod
    def params(cls) -> AWXParams:
        if cls._params is None:
            cls._params = AWXParams(**cls.config.params)
        return cls._params

    def __init__(self, client: Client):
        # polling here - polling of input requests
        # request timeout - timeout for API request
        super().__init__(client,
                         polling_interval=.1,  # polling for command inputs
                         request_timeout=10,
                         ignore_duplicates=False)

    async def start(self, stop_event: asyncio.Event):
        await self.client.start_consuming()
        await self.client.subscribe(self.params().topic_in)
        await super().start(stop_event)

    async def get_templates(self, filters: dict = None) -> Optional[list]:
        """Returns list of all job templates for csm organization

        This methods support Ansible tower filtering
        https://docs.ansible.com/ansible-tower/latest/html/towerapi/filtering.html
        """
        async with get_session() as session:
            async with session.get(self.params().host + '/job_templates', params=filters, timeout=5) as response:
                response_data = await response.json()
            assert response.status == 200, f'Expected response 200, got {response.status}'
        try:
            return _status_json(response_data['results'])
        except KeyError:
            LOGGER.error('No `results` field found in /job_templates response: \nResponse: %s', response_data)
            raise


def single_template_filter(template_name: str):
    """Filter by exact template name"""
    if template_name:
        return {'name__iexact': template_name}
    return None


def _status_json(json_data) -> List[TemplateStatus]:
    """Get status for all templates or for concrete template"""
    statuses = []
    for template_data in json_data:
        statuses.append(
            TemplateStatus(name=template_data['name'],
                           last_run_timestamp=template_data['last_job_run'],
                           last_status=template_data['status'],
                           playbook=template_data['playbook']))
    return statuses
